#include "MpdNuclei.h"

ClassImp(MpdNuclei);

/// Default constructor
MpdNuclei::MpdNuclei(){
}

/** Constructor.
  \param name Task name
  \param outputName Output file name (histograms will be stored here)
  \param settings_filename Full path to the MpdNuclei wagon configuration file (JSON formatted)
*/
MpdNuclei::MpdNuclei(const char *name, const char *outputName, const char* settings_filename) : MpdAnalysisTask(name, outputName){
  mParamConfig = outputName;
  settings_file = settings_filename;
}


/** Initialization: read settings from the JSON-file, book histograms.

   Histograms will be created automatically for each:
   - particle
   - centrality bin
   - PID method (MC PDG, MpdPid or evPID wagon)
*/
void MpdNuclei::UserInit(){
  printf("MpdNuclei::Initialization\n");

  read_settings_json(settings_file);           // Read settings file (JSON-formatted)
  if(s__gl_Verbose) print_settings();          // Print settings if s__gl_Verbose is set to true
  n_centrality_bins = s__ev_Centrality.size(); // Number of centrality bins taken from the settings file

  if(s__gl_PID == 2){                   // Initialization of the MpdPid (see its documentation)
    fPID = new MpdPid(s__id_TOFSigma, s__id_TPCSigma, s__mid_Energy, s__mid_Coef, s__mid_Generator.c_str(), s__mid_Tracking.c_str(), s__mid_IniString.c_str());
  }

  fOutputList = new TList();            // Standard list for the output histograms (MpdAnalysisTask)
  fOutputList -> SetOwner(kTRUE);       // Some default line (MpdAnalysisTask)
  TH1::AddDirectory(kFALSE);            // Some default line (MpdAnalysisTask)
  TH2::AddDirectory(kFALSE);            // Some default line (MpdAnalysisTask)

  TH1I *h1i__dummy = nullptr;           // Dummy histogram of the TH1I type, used later on in loops for histograms booking
  TH1F *h1d__dummy = nullptr;           // Same for TH1F
  TH2F *h2d__dummy = nullptr;           // Same for TH2F
  std::vector<TH2F*> row__h2d;          // Vector of TH2F type histograms -- used as "row" later on for histograms booking in the each centrality bin
  TProfile2D *p2d__dummy  = nullptr;    // Dummy 2D profile for the "pT_reco - pT_mc" profiles
  std::vector<TProfile2D*> row__p2d;    // Vector of the TProfile2D type for the "pT_reco - pT_mc" pT corrections
  TProfile   *p1d__dummy  = nullptr;    // Dummy 1D profile for the "pT_reco - pT_mc" profiles
  std::vector<TProfile*>   row__p1d;    // Vector of the TProfile type for the "pT_reco - pT_mc" pT corrections

  h__events = new TH1I("h__events", "Number of events for each centrality bin", n_centrality_bins, 0, n_centrality_bins+1);
  fOutputList -> Add(h__events);

// Technical histogram limits -- can be also moved to the JSON file, but this histograms are used just as an illustration
  const float dedx_lims[6] = {320, -4.0, 4.0, 1E4, 0.0, 5E2}; // dE/dx hists limits: nbinsx, xlow, xup, nbinsy, ylow, yup
  const float m2_lims[6] = {320, -4.0, 4.0, 2E2, 0.0, 10.0};  // m^2 histograms limits
  const float dcaxy_lims[3] = {80, 0, 20};                    // DCA X and Y limits
  const float dcaz_lims[3] = {600, 0, 150};                   // DCA Z limits
// --- common dE/dx histogram for all particle species and all centalities
  h__dedx = book_th2d("h__dedx", "dEdx vs P for all particles", "p*q, GeV/c", "dE/dx, arb. units", -1, -1, dedx_lims);
  h__dedx -> SetOption("colz");
  init_th2d(hv__dedx, "h__dedx", "dEdx vs P", "p*q, GeV/c", "dE/dx, arb. units", dedx_lims);

  h__m2 = book_th2d("h__m2", "m^{2} vs P for all particles", "p*q, GeV/c", "m^{2}, GeV^{2}/c^{4}", -1, -1, m2_lims);
  h__m2 -> SetOption("colz");
  init_th2d(hv__m2, "h__m2", "m^{2} vs P for all particles", "p*q, GeV/c", "m^{2}, GeV^{2}/c^{4}", m2_lims);

  init_th1d(hv__dca,      "h__dca",      "DCA",             "DCA, cm", "counts" , dcaxy_lims);
  init_th1d(hv__dca_sec,  "h__dca_sec",  "DCA (secondary)", "DCA, cm", "counts" , dcaxy_lims);
  init_th1d(hv__dca_prim, "h__dca_prim", "DCA (primary)",   "DCA, cm", "counts" , dcaxy_lims);

// --- histograms for particles identified by the MC PDG code
  if(s__gl_MC){
    init_phasespace(hv__pty_mc, "h__pty_mc", "p_{T} vs y (MC)"); // pT(y) histograms for selected particles and centrality bins
  } // if(s__gl_MC)


// --- histograms for particles identified by the evPID wagon or MpdPid
  if(s__gl_PID >= 0 && s__gl_PID < 3){
    if(s__gl_PID == 2){
      init_phasespace(hv__pty_pid, "h__pty_pid", "p_{T} vs y (PID)"); // pT(y) histograms for selected particles and centrality bins, combined PID
    }
    if(s__gl_PID == 0){
      init_phasespace(hv__pty_pid_tpc, "h__pty_pid_tpc", "p_{T} vs y (PID), evPID: TPC"); // pT(y) histograms for selected particles and centrality bins, evPID wagon, PID by TPC
      init_phasespace(hv__pty_pid_tof, "h__pty_pid_tof", "p_{T} vs y (PID), evPID: ToF"); // pT(y) histograms for selected particles and centrality bins, evPID wagon, PID by ToF
    }
    if(s__mid_dEdx) init_phasespace(hv__pty_pid_dedx, "h__pty_pid_dedx", "p_{T} vs y (PID, dE/dx only)"); // dE/dx only
    init_profile2d(pt__corr_out, "pt__corr", "p_{T} correction"); // vector of the TProfile2D for the pT corrections
    init_profile2d(pz__corr_out, "pz__corr", "p_{Z} correction"); // vector of the TProfile2D for the pZ corrections


    // TProfiles for the pT and pZ corrections -- input from the corrections file
    if(s__gl_ptcorr){
      for(int particle = 0; particle < s__p_List.size(); ++particle){
        const char* particle_name = s__p_List[particle].name;
        for(int c_bin = 0; c_bin < n_centrality_bins; ++c_bin){
          p2d__dummy = (TProfile2D*) s__gl_ptcorr_file -> Get(Form("p2d__ptcorr_%s_centrality%d", particle_name, c_bin));
          row__p2d.push_back(p2d__dummy);
        }
        pt__corr_2d.push_back(row__p2d); // 2D "pT_reco - pT_mc" profiles for the pT corrections -- load from the pT corrections file.
        row__p2d.clear();

        for(int c_bin = 0; c_bin < n_centrality_bins; ++c_bin){
          p2d__dummy = (TProfile2D*) s__gl_ptcorr_file -> Get(Form("p2d__pzcorr_%s_centrality%d", particle_name, c_bin));
          row__p2d.push_back(p2d__dummy);
        }
        pz__corr_2d.push_back(row__p2d);
       	row__p2d.clear();
      } // particles loop
    } // s__gl_ptcorr
  } // if(s__gl_PID)


// --- Efficiency and contamination histograms
  if(s__gl_Efficiency){
    init_phasespace(hv__eff_tpc_numerator,         "h__eff_tpc_numerator",         "TPC efficiency numerator");
    init_phasespace(hv__eff_tpc_denominator,       "h__eff_tpc_denominator",       "TPC efficiency denominator");
    init_phasespace(hv__cont_sec_numerator,        "h__cont_sec_numerator",        "Secondaries contamination numerator");
    init_phasespace(hv__cont_sec_denominator,      "h__cont_sec_denominator",      "Secondaries contamination denominator");
    init_phasespace(hv__eff_tof_denominator,       "h__eff_tof_denominator",       "ToF efficiency denominator");
    init_phasespace(hv__eff_pid_numerator,         "h__eff_pid_numerator",         "PID efficiency numerator");
    init_phasespace(hv__eff_pid_denominator,       "h__eff_pid_denominator",       "PID efficiency denominator");
    init_phasespace(hv__cont_pid_numerator,        "h__cont_pid_numerator",        "PID contamination numerator");
    init_phasespace(hv__cont_pid_denominator,      "h__cont_pid_denominator",      "PID contamination denominator");
    if(s__gl_PID == 2 && s__mid_dEdx){
      init_phasespace(hv__eff_pid_dedx_numerator,    "h__eff_pid_dedx_numerator",    "PID efficiency numerator (dE/dx only)");
      init_phasespace(hv__cont_pid_dedx_denominator, "h__cont_pid_dedx_denominator", "PID contamination denominator (dE/dx only)");
      init_phasespace(hv__cont_pid_dedx_numerator,   "h__cont_pid_dedx_numerator",   "PID contamination numerator (dE/dx only)");
    }
    if(s__gl_PID == 0){
      init_phasespace(hv__eff_pid_tpc_numerator,    "h__eff_pid_tpc_numerator",    "PID efficiency numerator (evPID TPC only)");
      init_phasespace(hv__cont_pid_tpc_denominator, "h__cont_pid_tpc_denominator", "PID contamination denominator (evPID TPC only)");
      init_phasespace(hv__cont_pid_tpc_numerator,   "h__cont_pid_tpc_numerator",   "PID contamination numerator (evPID TPC only)");

      init_phasespace(hv__eff_pid_tof_numerator,    "h__eff_pid_tof_numerator",    "PID efficiency numerator (evPID ToF only)");
      init_phasespace(hv__cont_pid_tof_denominator, "h__cont_pid_tof_denominator", "PID contamination denominator (evPID ToF only)");
      init_phasespace(hv__cont_pid_tof_numerator,   "h__cont_pid_tof_numerator",   "PID contamination numerator (evPID ToF only)");
    }
  }
} // MpdNuclei::UserInit()



/**
   Here each event is processed:
   - "event quality" checks are performed (via MpdNuclei::bad_event())
   - "track quality" checks are performed with the MpdNuclei::bad_track() subroutine
   - Centrality bin is selected via the MpdNuclei::select_centrality_bin() subroutine
   - Particle is identified (MC, MpdPid or evPID)
   - Histograms for the identified histograms are filled
  \param event MpdAnalysisEvent to process
*/
void MpdNuclei::ProcessEvent(MpdAnalysisEvent &event){
  if(bad_event(event)) return;                                             // Skip "bad" event
  int current_centrality_bin = select_centrality_bin(event.getCentrTPC()); // Centrality bin selection
  if(current_centrality_bin == -1) return;                                 // Skip event if the centrality bin can not be selected
  h__events -> AddBinContent(current_centrality_bin+1);                    // Increase the events counter for the selected centrality bin
                                                                           // All needed branches are accessed in this code block:
  mMpdGlobalTracks = event.fMPDEvent->GetGlobalTracks();                   //   MPD global tracks,
  mKalmanTracks    =              event.fTPCKalmanTrack;                   //   TPC Kalman tracks,
  mMCTracks        =                     event.fMCTrack;                   //   Monte-Carlo tracks,
  tofMatches       =                 event.fTOFMatching;                   //   ToF matching

  MpdTofMatchingData *match = new MpdTofMatchingData();                    // Yet another method for the ToF matching
  if(s__gl_TOF == 2){                                                      // Must be more robust than the checks
    int nTofMatch = tofMatches -> GetEntries();                            //   for the flags 2 and 6
    for(int itof = 0; itof < nTofMatch; ++itof){                           //   or some strange N-sigma checks
      match = (MpdTofMatchingData*) tofMatches -> UncheckedAt(itof);
      map_tof_matching[match -> GetKFTrackIndex()] = itof;
    }
  }

  int ntrmc = mMCTracks -> GetEntries();                                    // The number of the original Monte-Carlo tracks
  std::vector<int> is_splitted(ntrmc, -1);                                  // Initialization of the vector for the 'splitted' track removal

  // The Monte-Carlo loop
  if(s__gl_MC){                                                             // If s__gl_MC is set this code will collect the phase-spaces
    for (long int i = 0; i < ntrmc; i++) {                                  //   for the original primary Monte-Carlo particles
      MpdMCTrack* mctrack = (MpdMCTrack*) mMCTracks -> At(i);               // Get/open the MC track
      if(mctrack -> GetMotherId() != -1) continue;                          // If the MpdMCTrack::GetMotherId() == -1  -- the particle is primary
      int current_particle_mc = particle_by_pdg(mctrack -> GetPdgCode());   // Identify it by the PDG code (e.g. 2212 = p, -211 = pi minus etc)
      if(current_particle_mc == -1) continue;                               // Only particles defined in the settings file are selected to proceed further
      float pt_mc               = mctrack -> GetPt();                       // Get particle transverse momentum pT
      float rapidity_mc         = rapidity_mc = get_rapidity_mc(mctrack, s__p_List[current_particle_mc].mass); // Fix for the nuclei: rapidity MUST be computed 'by hands'!!!
      hv__pty_mc[current_particle_mc][current_centrality_bin] -> Fill(rapidity_mc, pt_mc);                     // Fill the MC phase space (pT-y) 2D histogram
      if(s__gl_Efficiency) hv__eff_tpc_denominator[current_particle_mc][current_centrality_bin] -> Fill(rapidity_mc, pt_mc); // Same for the TPC efficiency denominator
    }
  }


  // The Reco loop
  int ntr   = mKalmanTracks -> GetEntries();                                             // The number of the reconstructed tracks
  for (long int i = 0; i < ntr; i++) {
    MpdTrack          *mpdtrack = (MpdTrack *) mMpdGlobalTracks -> UncheckedAt(i);       // The Global track is accessed
    MpdTpcKalmanTrack *kftrack  = (MpdTpcKalmanTrack *) mKalmanTracks -> UncheckedAt(i); // The corresponding TPC Kalman track will also be used
    int mcId = kftrack -> GetTrackID();                                                  // MpdTpcKalmanTrack::GetTrackID() gives the ID of the corresponding Monte Carlo track
    is_splitted[mcId] += 1;                                                              // Increase the 'splitting counter' for the current reconstructed KF track:
    if(is_splitted[mcId] >= 1) continue;                                                 //   if this track is 'splitted' (the counter > 1) -- skip the track
    MpdMCTrack* mctrack = (MpdMCTrack*) mMCTracks -> At(mcId);                           // Get the Monte Carlo track
    int   pdg                 = mctrack -> GetPdgCode();                                 // The particle Monte-Carlo PDG code
    int   prodId              = mctrack -> GetMotherId();                                // The particle 'primacy': -1 = the particle is primary, any other -- secondary
    int   current_particle_mc = particle_by_pdg(pdg);                                    // Particle position in the particles vector (read in the settings file)
    float dca = sqrt( pow(mpdtrack->GetDCAX(),2) + pow(mpdtrack->GetDCAY(),2) + pow(mpdtrack->GetDCAZ(),2) ); // DCA calculation for the technical histogram -- REMOVE in the future!
    if(current_particle_mc != -1 && mpdtrack -> GetNofHits() >= s__tr_NHits){
  	  if(prodId == -1){
        hv__dca_prim[current_particle_mc][current_centrality_bin] -> Fill(dca); // DCA for the truly primary particle identified by MC
      }
      else{
        hv__dca_sec[current_particle_mc][current_centrality_bin] -> Fill(dca);  // DCA for the secondary particle identified by MC
      }
    }

    if(bad_track(mpdtrack)) continue;                                                  // Common 'track quality' checks -- see the documentation
    if(current_particle_mc != -1) hv__dca[current_particle_mc][current_centrality_bin] -> Fill(dca); // DCA for the accepted ('primary') particle identified by MC
    bool tof = isTOF(mpdtrack, i);                                                     // Check if ToF matching is avalable for this track by one of the three methods,
                                                                                       //   please, see the isTOF() routine description
    int   kfcharge = kftrack -> Charge();                                              // The reconstructed track charge
    double p       =  kftrack -> Momentum3().Mag() * kfcharge;                         // The reconstructed track full momentum and
    double dedx    =  kftrack -> GetDedx();                                            //   the dE/dx information

    float rapidity_mc         = mctrack -> GetRapidity();                              // The Monte-Carlo particle rapidity (CAN BE WRONG!!!)
    float pt_mc               = mctrack -> GetPt();                                    // The Monte-Carlo particle transverse momentum
    float p_mc                = mctrack -> GetP();                                     // The Monte-Carlo particle full momentum
    float pz_mc               = mctrack -> GetPz();                                    // The Monte-Carlo particle momentum z-component
    if (current_particle_mc != -1) rapidity_mc = get_rapidity_mc(mctrack, s__p_List[current_particle_mc].mass); // We have to recalculate the particle rapidity
    h__dedx -> Fill(p, dedx);                                                          // Fill the energy loss (dE/dx v momentum) histogram
    if (tof) h__m2 -> Fill(p, mpdtrack -> GetTofMass2());                              // And the mass-squared (m^2 v momentum) histogram if the ToF matching is available


    // Particle ID by MC, this code block is needed for the efficiencies calculation
    if(s__gl_MC && current_particle_mc != -1){
      if(s__gl_Efficiency){
        float pt_reco       = mpdtrack -> GetPt() * s__p_List[current_particle_mc].charge;  // Particle reconstructed transverse momentum 
        float pz_reco       = mpdtrack -> GetPz() * s__p_List[current_particle_mc].charge;  //   and momentum z-component must be multiplied by the number of charges:
                                                                                            //   (read in the settings file), e.g. He3 or He4 momentum must 
                                                                                            //   be multiplied by 2 as it has 2 charges (protons)
        float rapidity_reco = get_rapidity_reconstructed(mpdtrack, pt_reco, pz_reco, s__p_List[current_particle_mc].mass); // Recalculate the particle reconstructed rapidity

        // And then the reconstructed pT must be also corrected by the subtraction of the mean difference between the reconstructed and the Monte-Carlo momentums
        if(s__gl_ptcorr){
          pt_reco	      -= pt__corr_2d[current_particle_mc][current_centrality_bin] -> Interpolate(rapidity_reco, pt_reco);
          pz_reco       -= pz__corr_2d[current_particle_mc][current_centrality_bin] -> Interpolate(rapidity_reco, pz_reco);
          rapidity_reco  = get_rapidity_reconstructed(mpdtrack, pt_reco, pz_reco, s__p_List[current_particle_mc].mass); // Reconstructed rapidity is recalculated again
        }

        // The 'track quality' conditions are applied now to all tracks which are here.

        // The 'Secondaries contamination denominator', the 'ToF efficiency denominator' and the 'PID efficiency (dE/dx only) denominator'
        //   increase without any other special conditions
        hv__cont_sec_denominator[current_particle_mc][current_centrality_bin]  -> Fill(rapidity_mc, pt_mc);      // Secondaries contamination denominator
        hv__eff_tof_denominator[current_particle_mc][current_centrality_bin]   -> Fill(rapidity_reco, pt_reco);  // ToF efficiency denominator and PID efficiency (dE/dx only) denominator

        // The 'TPC efficiency numerator' increases if the particle is primary,
        //   otherwise the 'Secondaries contamination numerator' increases
        if(prodId == -1) hv__eff_tpc_numerator[current_particle_mc][current_centrality_bin] -> Fill(rapidity_mc, pt_mc);      // TPC efficiency numerator
        else             hv__cont_sec_numerator[current_particle_mc][current_centrality_bin] -> Fill(rapidity_reco, pt_reco); // Secondaries contamination numerator

        // If the ToF matching exists, the 'PID efficiency' and the 'ToF contamination denominator' increase,
        //   then, if thewe was an MC point in the ToF, the 'ToF efficiency numerator' increases,
        //   otherwise the 'ToF contamination numerator' increases.
        if(tof)hv__eff_pid_denominator[current_particle_mc][current_centrality_bin] -> Fill(rapidity_reco, pt_reco);  // PID efficiency denominator
                                                                                                                      // ToF efficiency numerator and ToF contamination denominator
      } // s__gl_Efficiency
    } // if(s__gl_MC)


    // Particle PID by the evPID wagon or the MpdPid class
    // This code block is needed for the efficiencies and the final results calculation
    int current_particle     = -1;
    int current_particle_tpc = -1;
    int current_particle_tof = -1;

    // Case for the MpdPid
    if(s__gl_PID == 2){
      bool res = 0;
      if(tof){                                          // If the ToF matching is avalable
        double m2 = mpdtrack -> GetTofMass2();          // Get the ToF m^2 from the global MPD track
        if (s__gl_TOF == 2) m2 = match -> GetMass2();   // If the 's__gl_TOF' is set to 2, then m^2 must be taken from the ToF matching branch
        res = fPID -> FillProbs(p*kfcharge, dedx, m2, kfcharge); // MpdPid combined identification by both dE/dx and mass-squared information
      }
      else{                                                  // If the ToF matching is not avalable we can try
        res = fPID -> FillProbs(p*kfcharge, dedx, kfcharge); //   to use only dE/dx information for the PID
      }
      if(!res) continue;                                        // If the MpdPid failed -- skip this and go to the next track
      current_particle = particle_by_pdg(fPID -> GetMaxProb()); // Otherwise -- get the PDG code of the 'most probable' particle via the MpdPid::GetMaxProb()
      if(current_particle == -1) continue;                      // Particle is not from the analysis list -- skip this and go to the next track
    }
    // end of s__gl_PID == 2
    // PID by the evPID wagon
    else if(s__gl_PID == 0){
      current_particle_tpc = particle_by_tpc(mpdtrack);                            // Identify the particle by the TPC dE/dx information
      if(tof){                                                                     // If the ToF matching exists,
        current_particle_tof = particle_by_tof(mpdtrack);                          //   identify the particle by the ToF mass-squared information
        if(current_particle_tpc == -1 && current_particle_tof == -1) continue;     //   if the particle can not be identified by both TPC and ToF
        if(current_particle_tpc != current_particle_tof) continue;                 //   or if TPC and ToF identification differs -- skip this and go to the next track
        if(veto_by_tof(mpdtrack, current_particle_tof)) current_particle_tof = -1; //   If the particle, identified by the ToF is vetoed by another particle --
      }                                                                            //     do not proceed the analysis for the evPID ToF PID case.
      if(veto_by_tpc(mpdtrack, current_particle_tpc)) current_particle_tpc = -1;   //   If the particle, identified by the TPC is vetoed by another particle --
    }                                                                              //      do not proceed the analysis for the evPID TPC PID case.
    // end of s__gl_PID == 0 
    else continue; // Any other s__gl_PID value -- skip the PID and go to the next track


    // At this moment the particle is identified!
    //      We can collect the final results

    // MpdPid dE/dx only case
    if(s__gl_PID == 2){
      float pt_reco       = mpdtrack -> GetPt() * s__p_List[current_particle].charge; // The reconstructed particle transverse momentum
      float pz_reco       = mpdtrack -> GetPz() * s__p_List[current_particle].charge; //   and the momentum z-component must be multiplied by the number of charges,
                                                                                      //   e.g. He3 or He4 momentum must be multiplied by 2 as it has 2 charges (protons)
      float rapidity_reco = get_rapidity_reconstructed(mpdtrack, pt_reco, pz_reco, s__p_List[current_particle].mass); // The reconstructed particle rapidity must be also recalculated
      // Correction of the reconstructed pT by the subtraction of the mean difference between the reconstructed and the Monte-Carlo momentums
      if(s__gl_ptcorr){
        pt_reco -= pt__corr_2d[current_particle][current_centrality_bin] -> Interpolate(rapidity_reco, pt_reco);
        pz_reco -= pz__corr_2d[current_particle][current_centrality_bin] -> Interpolate(rapidity_reco, pz_reco);
        rapidity_reco  = get_rapidity_reconstructed(mpdtrack, pt_reco, pz_reco, s__p_List[current_particle].mass); // Reconstructed rapidity is recalculated again
      }
      // This is the case, when only the dE/dx information is available (no ToF matching) for the 'MpdPid' case
      // The 's__mid_dEdx' parameter must be also set to 1 to get this code block works ('MpdPid.Add_dedx_only = 1').
      if(s__mid_dEdx && !tof){
        hv__dedx[current_particle][current_centrality_bin] -> Fill(p, dedx);                                       // Fill the technical dE/dx histogram for the identified particle
        hv__pty_pid_dedx[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco);                // The uncorrected phase-space for the identified particle (dE/dx PID only!)
        if(s__gl_Efficiency){
          // The 'dE/dx only PID contamination denominator' increases when particle is (somehow) identified
          hv__cont_pid_dedx_denominator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco); // PID contamination denominator for the dE/dx only case
          // The 'dE/dx only PID efficiency numerator' increases when the particle is identified correctly,
          //    otherwise, if the particle is identified wrongly, the 'dE/dx only PID contamination numerator' increases
          if(s__p_List[current_particle].pdg == pdg){
            hv__eff_pid_dedx_numerator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco);  // PID efficiency numerator for the dE/dx only case
          }
          else{
            hv__cont_pid_dedx_numerator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco); // PID contamination numerator for the dE/dx only case
          }
        } // s__gl_Efficiency
      } // End of the MpdPid dE/dx only case
      // MpdPid 'combined' case
      // This is the case, when the both dE/dx and m^2 information ('Combined PID') is available for the 'MpdPid' case
      if(tof){
        hv__m2[current_particle][current_centrality_bin] -> Fill(p, mpdtrack -> GetTofMass2());               // Fill the technical m^2 histogram for the identified particle
        hv__pty_pid[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco);                // The uncorrected phase-space for the identified particle
        if(s__gl_Efficiency){
          // The 'PID contamination denominator' increases when particle is (somehow) identified
          hv__cont_pid_denominator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco); // PID contamination denominator
          // The 'PID efficiency numerator' increases when the particle is identified correctly,
          //    otherwise, if the particle is identified wrongly, the 'PID contamination numerator' increases
          if(s__p_List[current_particle].pdg == pdg){
            hv__eff_pid_numerator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco);  // PID efficiency numerator
            // The difference between the reconstructed and the Monte-Carlo momentums is also filled when the particle is identified correctly
            pt__corr_out[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco, pt_reco - pt_mc); // These 2D prrofiles are needed for the pT
            pz__corr_out[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pz_reco, pz_reco - pz_mc); //   and the pZ correction
          }
          else{
            hv__cont_pid_numerator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco); // PID contamination numerator
          }
        } // s__gl_Efficiency
      } // tof
    } // End of the MpdPid 'combined' case
    // evPID dE/dx only case
    else if(s__gl_PID == 0){
      if(current_particle_tpc > -1){
        current_particle = current_particle_tpc;
        hv__dedx[current_particle][current_centrality_bin] -> Fill(p, dedx);            // Fill the technical dE/dx histogram for the identified particle
        float pt_reco       = mpdtrack -> GetPt() * s__p_List[current_particle].charge; // The reconstructed particle transverse momentum
        float pz_reco       = mpdtrack -> GetPz() * s__p_List[current_particle].charge; //   and the momentum z-component must be multiplied by the number of charges,
                                                                                        //   e.g. He3 or He4 momentum must be multiplied by 2 as it has 2 charges (protons)
        float rapidity_reco = get_rapidity_reconstructed(mpdtrack, pt_reco, pz_reco, s__p_List[current_particle].mass); // The reconstructed particle rapidity must be also recalculated
        // Correction of the reconstructed pT by the subtraction of the mean difference between the reconstructed and the Monte-Carlo momentums
        if(s__gl_ptcorr){
          pt_reco -= pt__corr_2d[current_particle][current_centrality_bin] -> Interpolate(rapidity_reco, pt_reco);
          pz_reco -= pz__corr_2d[current_particle][current_centrality_bin] -> Interpolate(rapidity_reco, pz_reco);
          rapidity_reco  = get_rapidity_reconstructed(mpdtrack, pt_reco, pz_reco, s__p_List[current_particle].mass); // Reconstructed rapidity is recalculated again
        }
        hv__pty_pid_tpc[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco);                // The uncorrected phase-space for the identified particle
        if(s__gl_Efficiency){
          // The 'TPC PID contamination denominator' increases when particle is (somehow) identified
          hv__cont_pid_tpc_denominator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco); // PID contamination denominator for the evPID TPC only case
          // The 'TPC PID efficiency numerator' increases when the particle is identified correctly,
          //    otherwise, if the particle is identified wrongly, the 'TPC PID contamination numerator' increases
          if(s__p_List[current_particle].pdg == pdg){
            hv__eff_pid_tpc_numerator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco);  // PID efficiency numerator for the evPID TPC only case
          }
          else{
            hv__cont_pid_tpc_numerator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco); // PID contamination numerator for the evPID TPC only case
          }
        } // s__gl_Efficiency
      } // End of the evPID dE/dx only case
      // evPID m^2 only case
      if(current_particle_tof > -1){ // evPID ToF only case
        current_particle = current_particle_tof;
        hv__m2[current_particle][current_centrality_bin] -> Fill(p, mpdtrack -> GetTofMass2()); // Fill the technical m^2 histogram for the identified particle
        float pt_reco       = mpdtrack -> GetPt() * s__p_List[current_particle].charge;         // The reconstructed particle transverse momentum
        float pz_reco       = mpdtrack -> GetPz() * s__p_List[current_particle].charge;         //   and the momentum z-component must be multiplied by the number of charges,
                                                                                                //   e.g. He3 or He4 momentum must be multiplied by 2 as it has 2 charges (protons)
        float rapidity_reco = get_rapidity_reconstructed(mpdtrack, pt_reco, pz_reco, s__p_List[current_particle].mass); // The reconstructed particle rapidity must be also recalculated
        // Correction of the reconstructed pT by the subtraction of the mean difference between the reconstructed and the Monte-Carlo momentums
        if(s__gl_ptcorr){
          pt_reco -= pt__corr_2d[current_particle][current_centrality_bin] -> Interpolate(rapidity_reco, pt_reco);
          pz_reco -= pz__corr_2d[current_particle][current_centrality_bin] -> Interpolate(rapidity_reco, pz_reco);
          rapidity_reco  = get_rapidity_reconstructed(mpdtrack, pt_reco, pz_reco, s__p_List[current_particle].mass); // Reconstructed rapidity is recalculated again
        }
        hv__pty_pid_tof[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco);                // The uncorrected phase-space for the identified particle
        if(s__gl_Efficiency){
          // The 'ToF PID contamination denominator' increases when particle is (somehow) identified
          hv__cont_pid_tof_denominator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco); // PID contamination denominator for the evPID ToF only case
          // The 'ToF PID efficiency numerator' increases when the particle is identified correctly,
          //    otherwise, if the particle is identified wrongly, the 'ToF PID contamination numerator' increases
          if(s__p_List[current_particle].pdg == pdg){
            hv__eff_pid_tof_numerator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco);  // PID efficiency numerator for the evPID ToF only case
          }
          else{
            hv__cont_pid_tof_numerator[current_particle][current_centrality_bin] -> Fill(rapidity_reco, pt_reco); // PID contamination numerator for the evPID ToF only case
          }
        } // s__gl_Efficiency
      } // End of the evPID m^2 only case
    } // evPID wagon s__gl_PID == 0 case
  } // i < ntr; The end of the loop over the MPD Global tracks

  map_tof_matching.clear(); // Clear the ToF matching information map for this event
  is_splitted.clear();      // Clear the track splitting vector for this event
} // MpdNuclei::ProcessEvent



/// Default destructor
MpdNuclei::~MpdNuclei(){}



/**
   Standard subroutine called at the end of the each wagon analysis.  
*/
void MpdNuclei::Finish(){
  if(s__gl_ptcorr) s__gl_ptcorr_file -> Close(); // Close the settings file
}



/** This subroutine performs the event quality checks.
  \param event MpdAnalysisEvent to analyse
  \return
    - true (1) if checks are not passed ("bad" event -- skip)  
    - false (0) if checks are passed ("good" event -- analyze)  

  The event can be skipped if:
    - The event vertex is not filled
    - The centrality is not defined
    - The event vertex is not reconstructed
    - The event vertex is beyond the limits set by MpdNuclei::s__ev_PrimaryVertexZ
*/
bool MpdNuclei::bad_event(MpdAnalysisEvent &event){
  if(!event.fVertex)                                        return true; // The event vertex is not filled -- skip event
  if(event.getCentrTPC() < 0 || event.getCentrTPC() >= 100) return true; // The centrality is not defined -- skip event
  MpdVertex *vertex = (MpdVertex *)event.fVertex -> First();             // Get the event primary vertex
  vertex -> Position(mPrimaryVertex);                                    // Access its position
  if(mPrimaryVertex.Z() == 0)                               return true; // The event vertex is not reconstructed -- skip event
  if(fabs(mPrimaryVertex.Z()) > s__ev_PrimaryVertexZ)       return true; // The event vertex is beyond the limits -- skip event
  return false;
}


/** This subroutine performs the track quality checks.
  \param track MpdTrack to analyse
  \return
    - true (1) if checks are not passed ("bad" track -- skip)  
    - false (0) if checks are passed ("good" track -- analyze)  

  The track can be skipped if:
    - The number of hits is less than MpdNuclei::s__tr_NHits
    - s__gl_DCA == 0 -- "classic" mode:
      - \f$ \sqrt{DCAX^{2} + DCAY^{2} + DCAZ^{2}} \f$ < MpdNuclei::s__tr_NSigmaDCAx
    - s__gl_DCA == 0 -- "NSigma" mode (the evPID wagon must be used first!!!):
      - The n-sigma for DCAx is larger than MpdNuclei::s__tr_NSigmaDCAx
      - The n-sigma for DCAy is larger than MpdNuclei::s__tr_NSigmaDCAy
      - The n-sigma for DCAz is larger than MpdNuclei::s__tr_NSigmaDCAz
    - The transverse momentum is less than MpdNuclei::s__tr_LowPtCut
*/
bool MpdNuclei::bad_track(MpdTrack* track){
  if (track -> GetNofHits() < s__tr_NHits)                 return true; // Number of hits cut
  if (s__gl_DCA == 0){
    if (sqrt( pow(track->GetDCAX(),2) + pow(track->GetDCAY(),2) + pow(track->GetDCAZ(),2) ) > s__tr_NSigmaDCAx) return true; // |DCAx| cut is used (in cm)
  }
  else if (s__gl_DCA == 1){
    if (fabs(track -> GetNSigmaDCAx()) > s__tr_NSigmaDCAx) return true; // |DCAx| cut in n-sigma
    if (fabs(track -> GetNSigmaDCAy()) > s__tr_NSigmaDCAy) return true; // |DCAy| cut in n-sigma
    if (fabs(track -> GetNSigmaDCAz()) > s__tr_NSigmaDCAz) return true; // |DCAz| cut in n-sigma
  }
  if (fabs(track->GetPt()) < s__tr_LowPtCut)             return true;   // Low transverse momentum cut
  return false;
}


/** This subroutine performs the ToF checks.

  \param track MpdTrack to analyse
  \param value KalmanTrack index
  \return 
    - true (1) if the ToF information exist;  
    - false (0) if the ToF information does not exist;  
  
  The ToF information exist if:
    - s__gl_TOF == 0 -- "classic" mode:
      - **MpdTrack::GetTofFlag()** is equeal to 2 or 6
    - s__gl_TOF == 1 -- "NSigma" mode:
      - The n-sigma **MpdTrack::GetTofDphiSigma()** is less than MpdNuclei::s__id_TOFDphiSigma
      - The n-sigma **MpdTrack::GetTofDzSigma()** is less than MpdNuclei::s__id_TOFDzSigma
    - s__gl_TOF == 2 -- "MpdTofMatchingData" mode:
      - MpdNuclei::map_tof_matching[MpdTofMatchingData::GetKFTrackIndex()] is filled
*/
bool MpdNuclei::isTOF(MpdTrack* track, const int value){
  if(s__gl_TOF == 0){
    if(track->GetTofFlag() == 2 || track->GetTofFlag() == 6) return true;
  }
  else if(s__gl_TOF == 1){
    if (fabs(track->GetTofDphiSigma()) < s__id_TOFDphiSigma
    && fabs(track->GetTofDzSigma())   <   s__id_TOFDzSigma) return true;
  }
  else if(s__gl_TOF == 2){
    if(map_tof_matching.count(value) > 0) return true;
  }
  return false;
}


/** Calculate the number of the centrality bin according to the TPC centrality.
  \return
    - the centrality bin number (integer) if the value of "getCentrTPC()"
     is within centrality bins limits defined in the settings file  
    - -1 in any other cases
  \param value TPC centrality value (MpdAnalysisEvent::getCentrTPC())
*/
int MpdNuclei::select_centrality_bin(const float value){
  for(int centrality_bin = 0; centrality_bin < n_centrality_bins; ++centrality_bin){
    if(value >=  s__ev_Centrality[centrality_bin].low &&
       value < s__ev_Centrality[centrality_bin].high) return centrality_bin;
  }
  return -1;
}


/** Particle identification via the Monte-Carlo PDG code
  \param value particle [Monte Carlo code](https://pdg.lbl.gov/2022/reviews/rpp2022-rev-monte-carlo-numbering.pdf)

  \return  
    - particle id after PID by the Monte-Carlo PDG code  
    - -1 in any other cases

  The "particle id" here is the particle position in the vector of particles defined in the settings file
*/
int MpdNuclei::particle_by_pdg(const int value){
  for(int particle = 0; particle < s__p_List.size(); ++particle){
    if (value == s__p_List[particle].pdg) return particle;
  }
  return -1;
}



/** Particle identification via the TPC dE/dx ("competitive" PID mode).
  \param track MpdTrack
  \return
    - particle id after TPC PID by the N-Sigma method of the evPID wagon
    - -1 in any other cases

    "particle id" here is the particle position in the vector of particles
    defined in the settings file.

  The particle is identified by the TPC if:
    - The n-sigma **MpdTrack::GetTPCNSigma()** is less than MpdNuclei::s__id_TPCSigma
    - The particle TPC n-sigma is the smaller within the particle list defined in the wagon settings file.
*/
int MpdNuclei::particle_by_tpc(MpdTrack* track){
  int particle_id = -1;
  int counter = 0;     // counter for the number of the nsigma entries for this track
  for(int particle = 0; particle < s__p_List.size(); ++particle){
    float TPC_sigma = fabs(track->GetTPCNSigma(s__p_List[particle].enum_position));
    if(TPC_sigma > s__id_TPCSigma) continue;
    if(signbit(track->GetCharge()) != signbit(s__p_List[particle].pdg)) continue;
    ++counter;
    if(counter > 1){ // Several particles passes the nSigma cut
      return -1;     // Do not accept this track -- particle can not be identified
    }
    particle_id = particle;
  }
  return particle_id;
}



/** This subroutine performs checks for the veto TPC (evPID).
  \param track MpdTrack to analyse
  \param suspect The particle id
  \return 
    - true (1) there a some other particles within the number of sigma MpdNuclei::s__eid_vetoTPC
    - false (0) if the selected particle is the only one particle within the MpdNuclei::s__eid_vetoTPC region
*/
bool MpdNuclei::veto_by_tpc(MpdTrack* track, const int suspect){
  for(int particle = 0; particle < s__p_List.size(); ++particle){
    if(s__p_List[particle].enum_position == s__p_List[suspect].enum_position) continue; // Do not make checks for the same type of particles

    float TPC_sigma = fabs(track->GetTPCNSigma(s__p_List[particle].enum_position));
    if(TPC_sigma <= s__eid_vetoTPC) return true;
  }
  return false;
}



/** Particle identification via the ToF mass-square ("competitive" PID mode).
  \param track MpdTrack
  \return
    - particle id after ToF PID by the N-Sigma method of the evPID wagon
    - -1 in any other cases

    "particle id" here is the particle position in the vector of particles
    defined in the settings file.

  The particle is identified by the ToF if:
    - The n-sigma **MpdTrack::GetTOFNSigma()** is less than MpdNuclei::s__id_TOFSigma
    - The particle ToF n-sigma is the smaller within the particle list defined in the wagon settings file.
*/
int MpdNuclei::particle_by_tof(MpdTrack* track){
  int particle_id = -1;
  int counter = 0;     // counter for the number of the nsigma entries for this track
  for(int particle = 0; particle < s__p_List.size(); ++particle){
    float TOF_sigma = fabs(track->GetTOFNSigma(s__p_List[particle].enum_position));
    if(TOF_sigma > s__id_TOFSigma) continue;
    if(signbit(track->GetCharge()) != signbit(s__p_List[particle].pdg)) continue;
    ++counter;
    if(counter > 1){ // Several particles passes the nSigma cut
      return -1;     // Do not accept this track -- particle can not be identified
    }
    particle_id = particle;
  }
  return particle_id;
}



/** This subroutine performs checks for the veto ToF (evPID).
  \param track MpdTrack to analyse
  \param suspect The particle id

  \return 
    - true (1) there a some other particles within the number of sigma MpdNuclei::s__eid_vetoToF
    - false (0) if the selected particle is the only one particle within the MpdNuclei::s__eid_vetoToF region
*/
bool MpdNuclei::veto_by_tof(MpdTrack* track, const int suspect){
  for(int particle = 0; particle < s__p_List.size(); ++particle){
    if(s__p_List[particle].enum_position == s__p_List[suspect].enum_position) continue; // Do not make checks for the same type of particles
    float TOF_sigma = fabs(track->GetTOFNSigma(s__p_List[particle].enum_position));
    if(TOF_sigma <= s__eid_vetoToF) return true;
  }
  return false;
}



/** Rapidity calculation for the Monte-Carlo track.
  \param track MpdMCTrack
  \param m Particle mass
  \return The calculated particle rapidity
*/
float MpdNuclei::get_rapidity_mc(MpdMCTrack* track, const float m){
  float px = track -> GetPx();
  float py = track -> GetPy();
  float pz = track -> GetPz();
  float e = sqrt(px*px + py*py + pz*pz + m*m);
  return 0.5 * log((e+pz)/(e-pz));
}


/** Rapidity calculation for the reconstructed track.
  \param track MpdTrack
  \param pt Particle transverse momentum
  \param pz Particle momentum z-component
  \param m Particle mass
  \return The calculated particle rapidity
*/
float MpdNuclei::get_rapidity_reconstructed(MpdTrack* track, const float pt, const float pz, const float m){
  float e  = sqrt(pt*pt + pz*pz + m*m);
  return 0.5 * log((e+pz)/(e-pz));
}


/** This subroutine creates a new TH2F histogram for the particle phasespace.
  \param hist_name The histogram name
  \param title The histogram title
  \param particle Particle position in the particles vector
  \param c_bin Centrality bin number
  \return The TH2F histogram
  The booked histogram is also added to the MpdAnalysisTask::fOutputList.
*/
TH2F* MpdNuclei::book_phasespace(const char *hist_name, const char *title, const int particle, const int c_bin){
  TH2F *hist = new TH2F(Form("%s_%s_centrality%d", hist_name, s__p_List[particle].name.Data(), c_bin),
                        Form("%s, %s, centrality bin %d; y; p_{T}, GeV/c", title, s__p_List[particle].name.Data(), c_bin),
                        (int)s__p_List[particle].rapidity_bins[0], s__p_List[particle].rapidity_bins[1], s__p_List[particle].rapidity_bins[2],
                        (int)s__p_List[particle].pt_bins[0],	 s__p_List[particle].pt_bins[1],       s__p_List[particle].pt_bins[2]);
  hist -> SetOption("colz");
  fOutputList -> Add(hist);
  return hist;
}


/** This routine initializes the vector of the TH2F histogram for the particle phase space.
  \param phasespace The vector of the phase spaces for each particle and each centrality
  \param hname Common name for the histograms in this vector
  \param htitle Common title for the histograms in this vector
*/
void MpdNuclei::init_phasespace(std::vector<std::vector<TH2F*>> &phasespace, const char *hname, const char* htitle){
  TH2F *h2d__dummy = nullptr;
  std::vector<TH2F*> row__h2d;
  for(int particle = 0; particle < s__p_List.size(); ++particle){
    for(int c_bin = 0; c_bin < s__ev_Centrality.size(); ++c_bin){
      h2d__dummy = book_phasespace(hname, htitle , particle, c_bin);
      row__h2d.push_back(h2d__dummy);
    }
    phasespace.push_back(row__h2d);
    row__h2d.clear();
  }
}



/** This subroutine creates a new TProfile2D
  \param hist_name The TProfile name
  \param title The TProfile title
  \param particle Particle position in the particles vector
  \param c_bin Centrality bin number
  \return The TProfile2D profile
  The booked profile is also added to the MpdAnalysisTask::fOutputList.
*/
TProfile2D* MpdNuclei::book_profile2d(const char *hist_name, const char *title, const int particle, const int c_bin){
  TProfile2D *hist = new TProfile2D(Form("%s_%s_centrality%d", hist_name, s__p_List[particle].name.Data(), c_bin),
                       	Form("%s, %s, centrality bin %d; y; RC, GeV/c; RC - MC, GeV/c", title, s__p_List[particle].name.Data(), c_bin),
                       	(int)s__p_List[particle].rapidity_bins[0], s__p_List[particle].rapidity_bins[1], s__p_List[particle].rapidity_bins[2],
                       	(int)s__p_List[particle].pt_bins[0],   s__p_List[particle].pt_bins[1],       s__p_List[particle].pt_bins[2]);
  hist -> SetOption("surf4");
  fOutputList -> Add(hist);
  return hist;
}



/** This routine initializes the vector of the TProfile2D
  \param vec The vector of the TProfiles2D* for each particle and each centrality
  \param hname Common name for the TProfiles in this vector
  \param htitle Common title for the TProfiles in this vector
*/
void MpdNuclei::init_profile2d(std::vector<std::vector<TProfile2D*>> &vec, const char *hname, const char* htitle){
  TProfile2D *dummy = nullptr;
  std::vector<TProfile2D*> row;
  for(int particle = 0; particle < s__p_List.size(); ++particle){
    for(int c_bin = 0; c_bin < s__ev_Centrality.size(); ++c_bin){
      dummy = book_profile2d(hname, htitle , particle, c_bin);
      row.push_back(dummy);
    }
    vec.push_back(row);
    row.clear();
  }
}



/** This subroutine creates a new TH2F object
  \param hist_name The TH2F name
  \param title The TH2F title
  \param xtitle The TH2F X-axis title
  \param ytitle The TH2F Y-axis title
  \param particle Particle position in the particles vector
  \param c_bin Centrality bin number
  \param lims The set of the TH2F histogram init parameters:
      - the number of x bins
      - the low x limit
      - the high x limit
      - the number of y bins
      - the low y limit
      - the high y limit
  \return The TH2F histogram
  The booked histogram is also added to the MpdAnalysisTask::fOutputList.
*/
TH2F* MpdNuclei::book_th2d(const char *hist_name, const char *title, const char *xtitle, const char *ytitle, const int particle, const int c_bin, const float lims[6]){
  const char *hname = hist_name;
  const char *htitle = Form("%s; %s; %s", title, xtitle, ytitle);
  if(particle > -1 && c_bin > -1){
    hname = Form("%s_%s_centrality%d", hist_name, s__p_List[particle].name.Data(), c_bin);
    htitle = Form("%s, %s, centrality bin %d; %s; %s", title, s__p_List[particle].name.Data(), c_bin, xtitle, ytitle);
  }
  TH2F *hist = new TH2F(hname, htitle, (int)lims[0], lims[1], lims[2], (int)lims[3], lims[4], lims[5]);
  fOutputList -> Add(hist);
  return hist;
}



/** This routine initializes the vector of the TH2F histograms
  \param vec The vector of the phase spaces for each particle and each centrality
  \param hname Common name for the histograms in this vector
  \param htitle Common title for the histograms in this vector
  \param xtitle The TH2F X-axis title
  \param ytitle The TH2F Y-axis title
  \param lims The set of the TH2F histogram init parameters:
      - the number of x bins
      - the low x limit
      - the high x limit
      - the number of y bins
      - the low y limit
      - the high y limit
*/
void MpdNuclei::init_th2d(std::vector<std::vector<TH2F*>> &vec, const char *hname, const char* htitle, const char *xtitle, const char *ytitle, const float lims[6]){
  TH2F *h2d__dummy = nullptr;
  std::vector<TH2F*> row__h2d;
  for(int particle = 0; particle < s__p_List.size(); ++particle){
    for(int c_bin = 0; c_bin < s__ev_Centrality.size(); ++c_bin){
      h2d__dummy = book_th2d(hname, htitle, xtitle, ytitle, particle, c_bin, lims);
      row__h2d.push_back(h2d__dummy);
    }
    vec.push_back(row__h2d);
    row__h2d.clear();
  }
}



/** This subroutine creates a new TH1F object
  \param hist_name The TH1F name
  \param title The TH1F title
  \param xtitle The TH1F X-axis title
  \param ytitle The TH1F Y-axis title
  \param particle Particle position in the particles vector
  \param c_bin Centrality bin number
  \param lims The set of the TH1F histogram init parameters:
      - the number of x bins
      - the low x limit
      - the high x limit
  \return The TH1F histogram
  The booked histogram is also added to the MpdAnalysisTask::fOutputList.
*/
TH1F* MpdNuclei::book_th1d(const char *hist_name, const char *title, const char *xtitle, const char *ytitle, const int particle, const int c_bin, const float lims[3]){
  const char *hname = hist_name;
  const char *htitle = Form("%s; %s; %s", title, xtitle, ytitle);
  if(particle > -1 && c_bin > -1){
    hname = Form("%s_%s_centrality%d", hist_name, s__p_List[particle].name.Data(), c_bin);
    htitle = Form("%s, %s, centrality bin %d; %s; %s", title, s__p_List[particle].name.Data(), c_bin, xtitle, ytitle);
  }
  TH1F *hist = new TH1F(hname, htitle, (int)lims[0], lims[1], lims[2]);
  fOutputList -> Add(hist);
  return hist;
}



/** This routine initializes the vector of the TH1F histograms
  \param vec The vector of the phase spaces for each particle and each centrality
  \param hname Common name for the histograms in this vector
  \param htitle Common title for the histograms in this vector
  \param xtitle The TH1F X-axis title
  \param ytitle The TH1F Y-axis title
  \param lims The set of the TH1F histogram init parameters:
      - the number of x bins
      - the low x limit
      - the high x limit
*/
void MpdNuclei::init_th1d(std::vector<std::vector<TH1F*>> &vec, const char *hname, const char* htitle, const char *xtitle, const char *ytitle, const float lims[3]){
  TH1F *dummy = nullptr;
  std::vector<TH1F*> row;
  for(int particle = 0; particle < s__p_List.size(); ++particle){
    for(int c_bin = 0; c_bin < s__ev_Centrality.size(); ++c_bin){
      dummy = book_th1d(hname, htitle, xtitle, ytitle, particle, c_bin, lims);
      row.push_back(dummy);
    }
    vec.push_back(row);
    row.clear();
  }
}



/**
  This subroutine reads the parameters from the
    JSON-formatted settings file
  \param fname Configuration file
*/
void MpdNuclei::read_settings_json(const char* fname){
  namespace settings = boost::property_tree;
  settings::ptree s_tree;
// Open settings file or terminate the program if it is not found
  try{
    settings::read_json(fname, s_tree);
  }
  catch(std::exception & e){
    printf("MpdNuclei::Fail:: %s\n", e.what());
    throw;
  }

// Global settings
  s__gl_Verbose    =              s_tree.get<bool>("Verbose", 1);
  s__gl_nMpdPID    = s_tree.get<short>("N_MPD_PID_Particles", 8);
  s__gl_MC         =              s_tree.get<bool>("make_MC", 1);
  s__gl_Efficiency =      s_tree.get<bool>("make_Efficiency", 1);
  s__gl_PID        =            s_tree.get<short>("PID_mode", 2);
  s__gl_DCA        =            s_tree.get<short>("DCA_mode", 0);
  s__gl_TOF        =            s_tree.get<short>("TOF_mode", 0);
  s__gl_ptcorr     =   s_tree.get<bool>("use_pt_corrections", 0);
  if(s__gl_ptcorr){
    std::string fcorrname = s_tree.get<std::string>("pt_corrections_file", "pt_corrections.root");
    s__gl_ptcorr_file = new TFile(fcorrname.c_str(), "READ");
    if(!s__gl_ptcorr_file || s__gl_ptcorr_file -> IsZombie()){
      printf("Error opening file: %s\n", fcorrname.c_str());
      exit(-1);
    }
  }

  if(!s__gl_MC && s__gl_Efficiency){
    printf("The efficiencies calculation can be enabled only if the 'make_MC' option is enabled\n");
    printf("Disabling the efficiencies calculation\n");
    s__gl_Efficiency = 0;
  }

  if((s__gl_PID < 0 || s__gl_PID > 2) && s__gl_Efficiency){
    printf("The efficiencies calculation can be enabled only if PID mode is within 0..2\n");
    printf("Disabling the efficiencies calculation\n");
    s__gl_Efficiency = 0;
  }

// MpdPid settings
  if(s__gl_PID == 2){
    if(!s_tree.get_child_optional("MpdPid")){
      printf("MpdPid settings are not defined\n");
      exit(-1);
    }
    s__mid_Energy     = s_tree.get<float>("MpdPid.Energy");
    s__mid_Coef       = s_tree.get<float>("MpdPid.Coef");
    s__mid_Generator  = s_tree.get<std::string>("MpdPid.Generator");
    s__mid_Tracking   = s_tree.get<std::string>("MpdPid.Tracking");
    s__mid_IniString  = s_tree.get<std::string>("MpdPid.IniString");
    s__mid_dEdx       = s_tree.get<bool>("MpdPid.Add_dedx_only");
  }
// evPID settings
  else if(s__gl_PID == 0){
    if(!s_tree.get_child_optional("evPID")){
      printf("evPID settings are not defined\n");
      exit(-1);
    }
    s__eid_vetoTPC     = s_tree.get<float>("evPID.VetoTPC");
    s__eid_vetoToF     = s_tree.get<float>("evPID.VetoToF");
  }

// Event settings
  if(!s_tree.get_child_optional("Events")){
    printf("Event cuts are not defined\n");
    exit(-1);
  }
  s__ev_PrimaryVertexZ = s_tree.get<double>("Events.PrimaryVertexZ");
  for(auto& centrality_array : s_tree.get_child("Events.Centrality")){
    float low, high;
    float* const elements[2] = {&low, &high};
    auto element = std::begin(elements);
    for (auto& node : centrality_array.second){
      **element++ = node.second.get_value<double>();
      if (element == std::end(elements)) break;
    }
    s__ev_Centrality.push_back({low, high});
  }

// Track settings
  if(!s_tree.get_child_optional("Tracks")){
    printf("Track cuts are not defined\n");
    exit(-1);
  }
  s__tr_NHits          = s_tree.get<int>("Tracks.NHits");
  s__tr_NSigmaDCAx     = s_tree.get<float>("Tracks.NSigmaDCAx");
  s__tr_NSigmaDCAy     = s_tree.get<float>("Tracks.NSigmaDCAy");
  s__tr_NSigmaDCAz     = s_tree.get<float>("Tracks.NSigmaDCAz");
  s__tr_LowPtCut       = s_tree.get<float>("Tracks.LowPtCut");

// PID settings
  if(!s_tree.get_child_optional("PID")){
    printf("PID cuts are not defined\n");
    exit(-1);
  }
  s__id_TPCSigma     = s_tree.get<float>("PID.TPCSigma");
  s__id_TOFSigma     = s_tree.get<float>("PID.TOFSigma");
  s__id_TOFDphiSigma = s_tree.get<float>("PID.TOFDphiSigma");
  s__id_TOFDzSigma   = s_tree.get<float>("PID.TOFDzSigma");


// Particle settings
  if(!s_tree.get_child_optional("Particles")){
    printf("Particles are not defined\n");
    exit(-1);
  }
  for(auto& particles_array : s_tree.get_child("Particles")){
    particle_info p;
    p.name              = particles_array.first.c_str();
    p.pdg               = particles_array.second.get_child("PDG").get_value<int>();
    p.mass              = particles_array.second.get_child("Mass").get_value<float>();
    p.charge            = particles_array.second.get_child("Charge").get_value<int>();
    p.enum_position     = particles_array.second.get_child("Enum").get_value<short>();

    auto node_pt = particles_array.second.get_child("pt_bins");
    for (auto it = node_pt.begin(); it != node_pt.end(); ++it){
      int position = std::distance(node_pt.begin(), it);
      p.pt_bins[position] = it->second.get_value<float>();
    }

    auto node_y = particles_array.second.get_child("rapidity_bins");
    for (auto it = node_y.begin(); it != node_y.end(); ++it){
      int position = std::distance(node_y.begin(), it);
      p.rapidity_bins[position] = it->second.get_value<float>();
    }

    s__p_List.push_back(p);
  }
} // MpdNuclei::read_settings_json()



/**
    This subroutine prints the wagon settings in a highly readable formatted output.
    This output can be turned off by the MpdNuclei::s__gl_Verbose flag set to 0.
*/
void MpdNuclei::print_settings(){
  printf("%-49s %6d\n", "MpdNuclei::Settings::Verbose:",                          s__gl_Verbose);
  printf("%-49s %6d\n", "MpdNuclei::Settings::nMpdPID:",                          s__gl_nMpdPID);
  printf("%-49s %6d\n", "MpdNuclei::Settings::MC:",                                    s__gl_MC);
  printf("%-49s %6d\n", "MpdNuclei::Settings::Efficiency:",                    s__gl_Efficiency);
  printf("%-49s %6d\n", "MpdNuclei::Settings::PID mode:",                             s__gl_PID);
  printf("%-49s %6d\n", "MpdNuclei::Settings::DCA mode:",                             s__gl_DCA);
  printf("%-49s %6d\n", "MpdNuclei::Settings::TOF mode:",                             s__gl_TOF);
  printf("%-49s %6d\n", "MpdNuclei::Settings::Use pT corr:",                       s__gl_ptcorr);
  if(s__gl_ptcorr) printf("%-20s %s\n", "corrections file:",     s__gl_ptcorr_file -> GetName());

  printf("%-49s %5.2f\n", "MpdNuclei::Settings::Events.PrimaryVertexZ:",   s__ev_PrimaryVertexZ);
  for(int centrality_bin = 0; centrality_bin < s__ev_Centrality.size(); ++centrality_bin){
    printf("MpdNuclei::Settings::Events.Centrality[%d]: [%5.2f %5.2f]\n",
           centrality_bin, s__ev_Centrality[centrality_bin].low, s__ev_Centrality[centrality_bin].high);
  }
  printf("%-49s %6d\n",   "MpdNuclei::Settings::Tracks.Nhits:",                     s__tr_NHits);
  printf("%-49s %6.2f\n", "MpdNuclei::Settings::Tracks.NSigmaDCAx:",           s__tr_NSigmaDCAx);
  printf("%-49s %6.2f\n", "MpdNuclei::Settings::Tracks.NSigmaDCAy:",           s__tr_NSigmaDCAy);
  printf("%-49s %6.2f\n", "MpdNuclei::Settings::Tracks.NSigmaDCAz:",           s__tr_NSigmaDCAz);
  printf("%-49s %6.2f\n", "MpdNuclei::Settings::Tracks.LowPtCut:",               s__tr_LowPtCut);
  printf("%-49s %6.2f\n", "MpdNuclei::Settings::PID.TPCSigma:",                  s__id_TPCSigma);
  printf("%-49s %6.2f\n", "MpdNuclei::Settings::PID.TOFSigma:",                  s__id_TOFSigma);
  printf("%-49s %6.2f\n", "MpdNuclei::Settings::PID.TOFDphiSigma:",          s__id_TOFDphiSigma);
  printf("%-49s %6.2f\n", "MpdNuclei::Settings::PID.TOFDzSigma:",              s__id_TOFDzSigma);
  if(s__gl_PID == 2){
    printf("%-49s %6.2f\n", "MpdNuclei::Settings::MpdPid.Energy:",                s__mid_Energy);
    printf("%-49s %6.2f\n", "MpdNuclei::Settings::MpdPid.Coef:",                    s__mid_Coef);
    printf("%-49s %6s\n",   "MpdNuclei::Settings::MpdPid.Generator:",  s__mid_Generator.c_str());
    printf("%-49s %6s\n",   "MpdNuclei::Settings::MpdPid.Tracking:",    s__mid_Tracking.c_str());
    printf("%-39s %16s\n",  "MpdNuclei::Settings::MpdPid.IniString:",  s__mid_IniString.c_str());
    printf("%-49s %6d\n",   "MpdNuclei::Settings::MpdPid.Add_dedx_only:",           s__mid_dEdx);
  }
  else if(s__gl_PID == 0){
    printf("%-49s %6.2f\n", "MpdNuclei::Settings::evPID.VetoTPC:",               s__eid_vetoTPC);
    printf("%-49s %6.2f\n", "MpdNuclei::Settings::evPID.VetoToF:",               s__eid_vetoToF);
  }

  for(int particle = 0; particle < s__p_List.size(); ++particle){
    printf("MpdNuclei::Settings::Particle[%d]: name=%s, pdg=%d, mass=%3.3f, charge=%2d, position in MpdPid enum: %d]\n",
           particle, s__p_List[particle].name.Data(), s__p_List[particle].pdg, s__p_List[particle].mass, s__p_List[particle].charge, s__p_List[particle].enum_position);
    printf("        pT bins: %4d %5.2f %5.2f\n",  (int)s__p_List[particle].pt_bins[0],  s__p_List[particle].pt_bins[1],  s__p_List[particle].pt_bins[2]);
    printf("  rapidity bins: %4d %5.2f %5.2f\n", (int)s__p_List[particle].rapidity_bins[0], s__p_List[particle].rapidity_bins[1], s__p_List[particle].rapidity_bins[2]);
  }
} // MpdNuclei::print_settings()
