#ifndef MPDNUCLEI_H
#define MPDNUCLEI_H

#include "MpdAnalysisTask.h"
#include "MpdTpcKalmanTrack.h"
#include "MpdKalmanFilter.h"
#include "MpdMCTrack.h"
#include "MpdVertex.h"
#include "MpdPid.h"
#include "MpdTofMatchingData.h"

#include "TFile.h"
#include "TF1.h"
#include "TH1.h"
#include "TH2.h"
#include "TH3.h"
#include "TProfile.h"
#include "TProfile2D.h"
#include "TString.h"

#include <string>

#define BOOST_BIND_GLOBAL_PLACEHOLDERS
#include <boost/foreach.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>

/**  \class MpdNuclei
  \ingroup physics
  MpdNuclei is a class for the nuclei analysis within MPDRoot "train" approach
  \author Viktar Kireyeu
  \date 03.09.2024
  \version 1.3
  \warning Under development
  \copyright Do not modify any line without the author's permission

[TOC]

# MpdNuclei wagon
MPDRoot wagon for the (light) nuclei analysis.

- [Dependencies](#dependencies)
  - [Branches](#branches)
  - [Wagons](#wagons)
- [Usage](#usage)
- [Configuration](#configuration)
  - [Global settings](#global-settings)
  - [Event settings](#event-settings)
  - [Track settings](#track-settings)
  - [PID settings](#pid-settings)
  - [MpdPid settings](#mpdpid-settings)
  - [evPID settings](#evpid-settings)
  - [Particles settings](#particles-settings)
  - [Example](#example)
- [Documentation](#documentation)

## Dependencies {#dependencies}
### Branches {#branches}
- MCTrack
- TpcKalmanTrack
- ZdcDigi
- Vertex
- MPDEvent
- TOFMatching
- MpdTofMatchingData (optional)  

### Wagons {#wagons}
- evCentrality
- evPID (optional)

## Usage {#usage}
Add these lines to your "train" macro (e.g. `RunAnalyses.C`):
```c
MpdNuclei taskNuclei("taskNuclei", "taskNuclei", "NucleiAna.json");
man.AddTask(&taskNuclei);
```

## Configuration {#configuration}
The JSON file format is used for the MpdNuclei wagon configuration.

### Global settings {#global-settings}
| Key | Default value | Description |
| ------------- |:-------------:| ------------- |
| Verbose | 1 | Print extra output |
| N_MPD_PID_Particles |8| Number of particles predefined for the PID in the `core/mpdBase/MpdTrack.h`|
| make_MC |0| Create and fill histograms for the particles identified by the MC information |
|make_Efficiency | 1 | Create and fill efficiency histograms |
|PID_mode|0|Switch for PID mode: 0 = "evPID"; 2 = "MpdPid"; any other value = turn off|
|DCA_mode|0|Switch for DCA mode: 0 = "classic"; 1 = "NSigma"|
|TOF_mode|0|Switch for ToF matching mode: 0 = "classic" -- check for the ToF flags 2 and 6; 1 = "NSigma" -- check for GetTofDphiSigma() and GetTofDzSigma(); 2 = use MpdTofMatchingData|
|use_pt_corrections | 0 | Switch for the creation of the pT correction 2D profiles |
|pt_corrections_file | N/A |  Name of the input file with the pT correction 2D profiles |
    
### Event settings {#event-settings}
There are no default values here.

| Key | Description |
| ------------- | ------------- |
| PrimaryVertexZ  | Cut for the z-coordinate of the the primary vertex |
| Centrality | A set of the centrality bins for to analyse (`[[min1, max1], [min2, max2], ...]`)|

### Track settings {#track-settings}
There are no default values here.

| Key | Description |
| ------------- | ------------- |
| NHits | Cut for the minimum number of hits |
| NSigmaDCAx | N-Sigma cut for the DCAx |
| NSigmaDCAy | N-Sigma cut for the DCAy |
| NSigmaDCAz | N-Sigma cut for the DCAz |
| LowPtCut | Cut for the minimum transverse momentum |

### PID settings {#pid-settings}
There are no default values here.

| Key | Description |
| ------------- | ------------- |
| TPCSigma | N-Sigma cut for the PID by the TPC dE/dx infomation |
| TOFSigma | N-Sigma cut for the PID by the ToF mass-squared information |
| TOFDphiSigma | N-Sigma cut for the ToF delta-phi (ToF-matching) |
| TOFDzSigma | N-Sigma cut for the ToF delta-z (ToF-matching) |

### MpdPid settings {#mpdpid-settings}
There are no default values here.
Plsease, see `core/mpdPid` for more explanation.

| Key | Description |
| ------------- | ------------- |
|s__mid_Energy|Collision energy|
|s__mid_Coef|MpdPid dE/dx coefficient|
|s__mid_Generator|MpdPid generator string|
|s__mid_Tracking|MpdPid tracking string|
|s__mid_IniString|MpdPid initialization sting|
|s__mid_dEdx|Switch for the use of the "dE/dx only" case in addition to the combined PID|

### evPID settings {#evpid-settings}
There are no default values here.

| Key | Description |
| ------------- | ------------- |
|s__eid_vetoTPC|N-sigma veto for the TPC pid|
|s__eid_vetoToF|N-sigma veto for the ToF pid|


### Particles settings {#particles-settings}
A separate block must be filled for the each particle. There is no default values.

| Key | Description |
| ------------- | ------------- |
| PDG | Monte Carlo particle code, [see the scheme](https://pdg.lbl.gov/2022/reviews/rpp2022-rev-monte-carlo-numbering.pdf) |
| Mass | Particle mass |
| Charge | The absolute value of the particle charge|
| Enum | Particle position in the `core/mpdBase/MpdTrack.h`|
|pt_bins| Number of pT bins, left limit, right limit |
|y_bins| Number of rapidity bins, left limit, right limit |


### Example {#example}
```json
{
    "Verbose": "1",
    "N_MPD_PID_Particles": "8",
    "make_MC": "1",
    "make_Efficiency": "1",
    "PID_mode": "2",
    "DCA_mode": "0",
    "TOF_mode": "0",
    "use_pt_corrections": "0",
    "pt_corrections_file": "pt_corrections.root",
    "Events": {
        "PrimaryVertexZ": "100",
        "Centrality": [[0, 20], [20, 40], [40, 80]]
    },
    "Tracks": {
        "NHits": "20",
        "NSigmaDCAx": "3",
        "NSigmaDCAy": "3",
        "NSigmaDCAz": "3",
        "LowPtCut": "0.05"
    },
    "PID": {
        "TPCSigma": "2",
        "TOFSigma": "2",
        "TOFDphiSigma": "3",
        "TOFDzSigma": "3"
    },
    "MpdPid": {
        "Energy": "9.2",
        "Coef": "1.0",
        "Generator": "PHQMD",
        "Tracking": "CFHM",
        "IniString": "pikaprdetrhe3he4",
        "Add_dedx_only": "1"
    },
    "evPID": {
        "VetoTPC": "3",
        "VetoToF": "3"
    },
    "Particles": {
        "pim": {
            "PDG": "-211",
            "Mass": "0.1395",
            "Charge": "1",
            "Enum": "1",
            "pt_bins": [60, 0.0, 6.0],
            "rapidity_bins": [60, -3.0, 3.0]
        },
        "pip": {
            "PDG": "211",
            "Mass": "0.1395",
            "Charge": "1",
            "Enum": "1",
            "pt_bins": [60, 0.0, 6.0],
            "rapidity_bins": [60, -3.0, 3.0]
        },
        "km": {
            "PDG": "-321",
            "Mass": "0.493677",
            "Charge": "1",
            "Enum": "2",
            "pt_bins": [60, 0.0, 6.0],
            "rapidity_bins": [60, -3.0, 3.0]
        },
        "kp": {
            "PDG": "321",
            "Mass": "0.493677",
            "Charge": "1",
            "Enum": "2",
            "pt_bins": [60, 0.0, 6.0],
            "rapidity_bins": [60, -3.0, 3.0]
        },
        "p": {
            "PDG": "2212",
            "Mass": "0.938",
            "Charge": "1",
            "Enum": "3",
            "pt_bins": [60, 0.0, 6.0],
            "rapidity_bins": [60, -3.0, 3.0]
        },
        "ap": {
            "PDG": "-2212",
            "Mass": "0.938",
            "Charge": "1",
            "Enum": "3",
            "pt_bins": [60, 0.0, 6.0],
            "rapidity_bins": [60, -3.0, 3.0]
        },
        "d": {
            "PDG": "1000010020",
            "Mass": "1.876",
            "Charge": "1",
            "Enum": "4",
            "pt_bins": [30, 0.0, 6.0],
            "rapidity_bins": [60, -3.0, 3.0]
        },
        "t": {
            "PDG": "1000010030",
            "Mass": "2.814",
            "Charge": "1",
            "Enum": "5",
            "pt_bins": [20, 0.0, 6.0],
            "rapidity_bins": [60, -3.0, 3.0]
        },
        "He3": {
            "PDG": "1000020030",
            "Mass": "2.814",
            "Charge": "2",
            "Enum": "6",
            "pt_bins": [20, 0.0, 6.0],
            "rapidity_bins": [60, -3.0, 3.0]
        },
        "He4": {
            "PDG": "1000020040",
            "Mass": "3.752",
            "Charge": "2",
            "Enum": "7",
            "pt_bins": [15, 0.0, 6.0],
            "rapidity_bins": [60, -3.0, 3.0]
        }
    }
}
```


## Documentation {#documentation}
To generate the MpdNuclei class documentation one can use the Doxygen genetator:
``` doxygen doxygen_config ```
The generated documentation will be placed in the newly created folder: ```html/index.html```.
*/
class MpdNuclei: public MpdAnalysisTask, public MpdPid {
  MpdPid  *fPID;

public:
  MpdNuclei();
  MpdNuclei(const char *name, const char *outputName = "taskName", const char* settings_filename = "file.json");
  ~MpdNuclei();

  void UserInit();
  void ProcessEvent(MpdAnalysisEvent &event);
  void Finish();

/** @struct centrality_bin
  A struct to handle the centrality bin limits
*/
  struct centrality_bin{
    float low;    ///< Low limit
    float high;   ///< High limit
  };

/** @struct particle_info
  A struct to handle the particle information from the
  configuration file.
*/
  struct particle_info{
    TString name;                ///< Particle name
    int     pdg;                 ///< Monte Carlo particle code, [see the scheme](https://pdg.lbl.gov/2022/reviews/rpp2022-rev-monte-carlo-numbering.pdf)
    float   mass;                ///< Particle mass
    int     charge;              ///< Particle charge
    short   enum_position;       ///< Particle position in the `core/mpdBase/MpdTrack.h`
    float   pt_bins[3];          ///< Number of pT bins, left limit, right limit.
    float   rapidity_bins[3];    ///< Number of rapidity bins, left limit, right limit.
  };


private:
  bool isInitialized = false;
  std::string mParamConfig;
  const char* settings_file;
  std::map<int,int> map_tof_matching;

/** @name Global settings
 */
///@{
  bool        s__gl_Verbose;     ///< Switch to handle an extra output.
  short       s__gl_nMpdPID;     ///< Number of particles predefined for the PID in the `core/mpdBase/MpdTrack.h`.
  bool        s__gl_MC;          ///< Switch for the creation of the histograms for the particles identified by the MC information.
  bool        s__gl_Efficiency;  ///< Switch for the creation of the efficiency histograms.
  short       s__gl_PID;         ///< Switch for PID mode: 0 = "evPID"; 2 = "MpdPid".
  short       s__gl_DCA;         ///< Switch for DCA mode: 0 = "classic"; 1 = "NSigma".
  short       s__gl_TOF;         ///< Switch for ToF matching mode:<br> 0 = "classic" -- check for the ToF flags 2 and 6;<br> 1 = "NSigma" -- check for MpdTrack::GetTofDphiSigma() and MpdTrack::GetTofDzSigma();<br> 2 = use MpdTofMatchingData
  bool        s__gl_ptcorr;      ///< Switch for the use of the pt corrections: 1 = 1D or 2 = 2D TPofiles
  TFile*      s__gl_ptcorr_file = nullptr;  ///< File with 2D profiles for the pT corrections
///@}
/** @name Event settings
 */
///@{
  float s__ev_PrimaryVertexZ;                   ///< Cut for the z-coordinate of the the primary vertex
  std::vector<centrality_bin> s__ev_Centrality; ///< Vector to store the centrality bins
  int n_centrality_bins;                        // The number of centrality bins
///@}
/** @name Track settings
 */
///@{
  int   s__tr_NHits;              ///< Cut for the minimum number of hits
  float s__tr_NSigmaDCAx;         ///< N-Sigma cut for the DCAx
  float s__tr_NSigmaDCAy;         ///< N-Sigma cut for the DCAy
  float s__tr_NSigmaDCAz;         ///< N-Sigma cut for the DCAz
  float s__tr_LowPtCut;           ///< Cut for the minimum transverse momentum
///@}
/** @name Commin PID settings
 */
///@{
  float s__id_TPCSigma;           ///< N-Sigma cut for the PID by the TPC dE/dx infomation
  float s__id_TOFSigma;           ///< N-Sigma cut for the PID by the ToF mass-squared information
  float s__id_TOFDphiSigma;       ///< N-Sigma cut for the matching to TOF
  float s__id_TOFDzSigma;         ///< N-Sigma cut for the matching to TOF
///@}
/** @name MpdPid settings
 */
///@{
  float s__mid_Energy;            ///< Collision energy
  float s__mid_Coef;              ///< MpdPid dE/dx coefficient
  std::string s__mid_Generator;   ///< MpdPid generator string
  std::string s__mid_Tracking;    ///< MpdPid tracking string
  std::string s__mid_IniString;   ///< MpdPid initialization sting
  bool s__mid_dEdx;               ///< Switch for the use of the MpdPid "dE/dx only" case in addition to the combined PID
///@}
/** @name evPID settings
 */
///@{
  float s__eid_vetoTPC;           ///< n-sigma veto for the TPC pid (evPID)
  float s__eid_vetoToF;           ///< n-sigma veto for the ToF pid (evPID)
///@}
/** @name Particle settings
 */
///@{
  std::vector<particle_info> s__p_List;    ///< Vector to store the particles information
///@}

  TVector3 mPrimaryVertex;
  TClonesArray *mMCTracks                       = nullptr;
  TClonesArray *mKalmanTracks                   = nullptr;
  TClonesArray *mMpdGlobalTracks                = nullptr;
  TClonesArray *tofMatches                      = nullptr;
/** @name Technical histograms which must be in a separate QA wagon
 */
///@{
  // DCA
  std::vector<std::vector<TH1F*>> hv__dca_sec;  ///< DCA histogram for the primary and secondary particle specie
  std::vector<std::vector<TH1F*>> hv__dca;      ///< DCA histogram for the accepted particle specie
  std::vector<std::vector<TH1F*>> hv__dca_prim; ///< DCA histogram for the particle specie primary by the MC information
  // dEdx
  TH2F *h__dedx = nullptr; ///< Common dE/dx histogram for all particle species
  std::vector<std::vector<TH2F*>> hv__dedx; ///< dE/dx histogram for the accepted particle specie
  // m2
  TH2F *h__m2 = nullptr; ///< Common dE/dx histogram for all particle species
  std::vector<std::vector<TH2F*>> hv__m2; ///< m2 histogram for the accepted particle specie
///@}
/** @name Identified particles phase-space histograms
 */
///@{
  std::vector<std::vector<TH2F*>> hv__pty_mc;       ///< Phase-space histograms for particles identified by the MC PDG code in the selected centrality bins
  std::vector<std::vector<TH2F*>> hv__pty_pid;      ///< Phase-space histograms for particles identified by the MpdPid or evPID wagon (N-Sigma method) in the selected centrality bins
  std::vector<std::vector<TH2F*>> hv__pty_pid_dedx; ///< ame but for the 'PID = dE/dx only' case
  std::vector<std::vector<TH2F*>> hv__pty_pid_tpc;  ///< Phase-space histograms for particles identified by the evPID wagon: TPC only
  std::vector<std::vector<TH2F*>> hv__pty_pid_tof;  ///< Phase-space histograms for particles identified by the evPID wagon: ToF only
///@}
/**  @name Efficiency histograms
 */
///@{
  std::vector<std::vector<TH2F*>> hv__eff_tpc_numerator;         ///< TPC efficiency numerator
  std::vector<std::vector<TH2F*>> hv__eff_tpc_denominator;       ///< TPC efficiency denominator
  std::vector<std::vector<TH2F*>> hv__cont_sec_numerator;        ///< Secondaries contamination numerator
  std::vector<std::vector<TH2F*>> hv__cont_sec_denominator;      ///< Secondaries contamination denominator
  std::vector<std::vector<TH2F*>> hv__eff_pid_numerator;         ///< PID efficiency numerator (MpdPid)
  std::vector<std::vector<TH2F*>> hv__eff_pid_denominator;       ///< PID efficiency denominator (MpdPid), ToF efficiency numerator and ToF contamination denominator
  std::vector<std::vector<TH2F*>> hv__cont_pid_numerator;        ///< PID contamination numerator (MpdPid)
  std::vector<std::vector<TH2F*>> hv__cont_pid_denominator;      ///< PID contamination denominator (MpdPid)
  std::vector<std::vector<TH2F*>> hv__eff_tof_denominator;       ///< ToF efficiency denominator
  std::vector<std::vector<TH2F*>> hv__eff_pid_dedx_numerator;    ///< PID efficiency numerator for the dE/dx only case (MpdPid)
  std::vector<std::vector<TH2F*>> hv__cont_pid_dedx_numerator;   ///< PID contamination numerator for the dE/dx only case (MpdPid)
  std::vector<std::vector<TH2F*>> hv__cont_pid_dedx_denominator; ///< PID contamination denominator for the dE/dx only case (MpdPid)
  std::vector<std::vector<TH2F*>> hv__eff_pid_tpc_numerator;     ///< PID efficiency numerator for the evPID TPC only case
  std::vector<std::vector<TH2F*>> hv__cont_pid_tpc_numerator;    ///< PID contamination numerator for the evPID TPC only case
  std::vector<std::vector<TH2F*>> hv__cont_pid_tpc_denominator;  ///< PID contamination denominator for the evPID TPC only case
  std::vector<std::vector<TH2F*>> hv__eff_pid_tof_numerator;     ///< PID efficiency numerator for the evPID ToF only case
  std::vector<std::vector<TH2F*>> hv__cont_pid_tof_numerator;    ///< PID contamination numerator for the evPID ToF only case
  std::vector<std::vector<TH2F*>> hv__cont_pid_tof_denominator;  ///< PID contamination denominator for the evPID ToF only case
///@}

  TH1I                                     *h__events;
  std::vector<std::vector<TProfile2D*>>  pt__corr_out;
  std::vector<std::vector<TProfile2D*>>  pz__corr_out;
  std::vector<std::vector<TProfile2D*>>   pt__corr_2d;
  std::vector<std::vector<TProfile*>>     pt__corr_1d;
  std::vector<std::vector<TProfile2D*>>   pz__corr_2d;

  bool  bad_event(MpdAnalysisEvent &event);
  bool  bad_track(MpdTrack* track);
  bool  isTOF(MpdTrack* track, const int value);
  int   select_centrality_bin(const float value);
  int   particle_by_pdg(const int value);
  int   particle_by_tpc(MpdTrack* track);
  bool  veto_by_tpc(MpdTrack* track, const int suspect);
  int   particle_by_tof(MpdTrack* track);
  bool  veto_by_tof(MpdTrack* track, const int suspect);
  float get_rapidity_mc(MpdMCTrack* track, const float m);
  float get_rapidity_reconstructed(MpdTrack* track, const float pt, const float pz, const float m);
  void  read_settings_json(const char* fname);
  void  print_settings();
  void  init_th1d(std::vector<std::vector<TH1F*>> &vec, const char *hname, const char* htitle, const char *xtitle, const char *ytitle, const float lims[3]);
  void  init_th2d(std::vector<std::vector<TH2F*>> &vec, const char *hname, const char* htitle, const char *xtitle, const char *ytitle, const float lims[6]);
  void  init_phasespace(std::vector<std::vector<TH2F*>> &phasespace, const char *hname, const char* htitle);
  void  init_profile2d(std::vector<std::vector<TProfile2D*>> &vec, const char *hname, const char* htitle);
  TH1F* book_th1d(const char *hist_name, const char *title, const char *xtitle, const char *ytitle, const int particle, const int c_bin, const float lims[3]);
  TH2F* book_th2d(const char *hist_name, const char *title, const char *xtitle, const char *ytitle, const int particle, const int c_bin, const float lims[6]);
  TH2F* book_phasespace(const char *hist_name, const char *title, const int particle, const int c_bin);
  TProfile2D* book_profile2d(const char *hist_name, const char *title, const int particle, const int c_bin);

  ClassDef(MpdNuclei, 1);
};
#endif
